#include "main.h"
#include <stdio.h>
#include <limits.h>


/* Compare wieghts to sort in DESCENDING order */
int CmpPkgNodeWt(const PKGLISTNODE *pnode1, const PKGLISTNODE *pnode2)
{
    if (pnode1->pkg.weight > pnode2->pkg.weight)
    {
        return -1; /* note! deliberately ulta for descending! */
    }
    else if (pnode1->pkg.weight == pnode2->pkg.weight)
    {
        return 0;
    }
    else
    {
        return 1;
    }
}

/* TODO: don't pick unreachable packages */

/* Note: return value could be null */
INTLISTNODE *WhatToPick(PKGLISTNODE *head)
{
    int len;
    PKGLISTNODE **pkgNodes = NULL;
    int i;
    PKGLISTNODE *node = head;
    int nowCapacity;
    INTLISTNODE *ret = NULL;

    /* Put the pointers to package nodes in an array */
    len = PkgListLen(head);

    pkgNodes = (PKGLISTNODE **)malloc(len * sizeof(PKGLISTNODE *));
    if (pkgNodes == NULL)
    {
        return NULL;
    }


    for (i = 0, node = head; i < len && node; i++, node = node->next)
    {
        pkgNodes[i] = node;
    }

    /* sort that array */
    qsort(pkgNodes, len, sizeof(PKGLISTNODE *), CmpPkgNodeWt);

    /* Get current carrying capacity */
    nowCapacity = gbMyCapacity - PkgListWtSum(gbMyPkgs);

    /* Pick whatever we can in descending order */
    ret = NULL;
    for (i = 0; i < len && nowCapacity > 0; i++)
    {
        PACKAGE pkg = pkgNodes[i]->pkg;

        if (pkg.weight <= nowCapacity &&
            Tile(pkg.destx, pkg.desty).pathVal != MAX_LEN)
        {
            INTLISTNODE *newNode = NewIntNode(pkg.id);
            if (newNode != NULL)
            {
                ret = AddIntNode(ret, newNode);
                nowCapacity -= pkg.weight;
            }
        }
    }

    free(pkgNodes);

    return ret;
}


void GetBidValue(long *bid, int *timeToDrop)
{
    float num;
    int foundWater, foundRobot;
    long bidVal;
    PKGLISTNODE *pkgn;

    *bid = 1;
    *timeToDrop = 0;

    /*** Are we in danger? ****/
    foundWater = 0;
    foundRobot = 0;

    /* west */
    if (Tile(gbMyX-1, gbMyY).robot != NULL)
    {
        foundRobot = 1;
    }
    if (Tile(gbMyX-1, gbMyY).type == WATER)
    {
        foundWater = 1;
    }
    /* east */
    if (Tile(gbMyX+1, gbMyY).robot != NULL)
    {
        foundRobot = 1;
    }
    if (Tile(gbMyX+1, gbMyY).type == WATER)
    {
        foundWater = 1;
    }
    /* north */
    if (Tile(gbMyX, gbMyY + 1).robot != NULL)
    {
        foundRobot = 1;
    }
    if (Tile(gbMyX, gbMyY+1).type == WATER)
    {
        foundWater = 1;
    }

    /* south */
    if (Tile(gbMyX, gbMyY-1).robot != NULL)
    {
        foundRobot = 1;
    }
    if (Tile(gbMyX, gbMyY-1).type == WATER)
    {
        foundWater = 1;
    }

    if (foundWater && foundRobot)
    {
        num = gbMyMoney;
        num *= DANGER_BID / 100.0;
        bidVal = num;

        if (bidVal < 1 || bidVal > gbMyMoney)
        {
            bidVal = 1;
        }

        *bid = bidVal;

        return;
    }

    /*** Is it time to drop? ***/
    pkgn = gbMyPkgs;

    while (pkgn)
    {
        if (pkgn->pkg.destx == gbMyX &&
            pkgn->pkg.desty == gbMyY)
        {
            bidVal = 1;

            if (foundRobot)
            {
                num = gbMyMoney;
                num *= DROP_BID / 100.0;
                bidVal = num;

                if (bidVal < 1 || bidVal > gbMyMoney)
                {
                    bidVal = 1;
                }
            }

            *bid = bidVal;
            *timeToDrop = 1;

            return;
        }

        pkgn = pkgn->next;
    }
}

void FindMinNeighbour(char *dir)
{
    long minVal = MAX_LEN;
    char minDir = NODIR;

    /* west */
    if (Tile(gbMyX-1, gbMyY).pathVal < minVal)
    {
        minVal = Tile(gbMyX-1, gbMyY).pathVal;
        minDir = WEST;
    }

    /* east */
    if (Tile(gbMyX+1, gbMyY).pathVal < minVal)
    {
        minVal = Tile(gbMyX+1, gbMyY).pathVal;
        minDir = EAST;
    }

    /* north */
    if (Tile(gbMyX, gbMyY + 1).pathVal < minVal)
    {
        minVal = Tile(gbMyX, gbMyY+1).pathVal;
        minDir = NORTH;
    }

    /* south */
    if (Tile(gbMyX, gbMyY-1).pathVal < minVal)
    {
        minVal = Tile(gbMyX, gbMyY-1).pathVal;
        minDir = SOUTH;
    }

    *dir = minDir;
}

void FindAnyNeighbour(char *dir)
{
    char dirs[4];
    int numDirs = 0;
    int randNum;

    /* west */
    if (Tile(gbMyX-1, gbMyY).type != WALL && Tile(gbMyX-1, gbMyY).type != WATER )
    {
        dirs[numDirs++] = WEST;
    }

    /* east */
    if (Tile(gbMyX+1, gbMyY).type != WALL && Tile(gbMyX+1, gbMyY).type != WATER )
    {
        dirs[numDirs++] = EAST;
    }

    /* north */
    if (Tile(gbMyX, gbMyY+1).type != WALL && Tile(gbMyX, gbMyY+1).type != WATER )
    {
        dirs[numDirs++] = NORTH;
    }

    /* south */
    if (Tile(gbMyX, gbMyY-1).type != WALL && Tile(gbMyX, gbMyY-1).type != WATER )
    {
        dirs[numDirs++] = SOUTH;
    }

    if (numDirs == 0)
    {
        *dir = NODIR;
    }

    randNum = rand() % numDirs;

    *dir = dirs[randNum];

    return;
}

void PackageDeliveryMode( long bid)
{
    long minVal;
    PKGLISTNODE *minPkg;
    PKGLISTNODE *p;
    char dir;

#ifdef ICFP_DEBUG
    printf("In package delivery mode\n");
#endif

    /* Package on my back with nearest dest */
    minVal = LONG_MAX;
    minPkg = NULL;

    p = gbMyPkgs;

    while ( p != NULL )
    {
        if ( Tile(p->pkg.destx, p->pkg.desty).pathVal < minVal )
        {
            minPkg = p;
            minVal = Tile(p->pkg.destx, p->pkg.desty).pathVal;
        }

        p = p->next;
    }

    /* If dest is right here, drop it */
    if ( minVal == 0 )
    {
        INTLISTNODE *dropPkgList = NULL;

        p = gbMyPkgs;

        while ( p != NULL )
        {
            if ( ( p->pkg.destx == gbMyX ) && ( p->pkg.desty == gbMyY ) )
            {
                INTLISTNODE *newNode = NewIntNode( p->pkg.id );
                if ( newNode != NULL )
                {
                    dropPkgList = AddIntNode( dropPkgList, newNode );
                }
            }
            p = p->next;
        }

        IssueDrop( bid, dropPkgList );
        FreeIntList( dropPkgList );
        return;
    }

    if ( minPkg )
    {
        /* Otherwise, see where to go next */
        ShortestPath( minPkg->pkg.destx, minPkg->pkg.desty );

        FindMinNeighbour( &dir );

        IssueMove( bid, dir );
    }
    else
    {
        IssuePick( 1, NULL );
    }
}

void PackagePickMode( long bid)
{
    long minVal;
    PKGLISTNODE *minPkg;
    PKGLISTNODE *p, *prevp;
    char dir;

#ifdef ICFP_DEBUG
    printf("In PackagePickMode\n");
#endif

    /* Find nearest package */
    minVal = LONG_MAX;
    minPkg = NULL;

    p = gbKnownPkgs;

    while ( p != NULL )
    {
        if ( Tile(p->pkg.x, p->pkg.y).pathVal < minVal )
        {
            minPkg = p;
            minVal = Tile(p->pkg.x, p->pkg.y).pathVal;
        }

        p = p->next;
    }

    /* If its supposed to be here, forget it since I never picked it */
    if ( minVal == 0 )
    {
        p = gbKnownPkgs;
        prevp = NULL;

        while ( p != NULL )
        {
            if ( Tile(p->pkg.x, p->pkg.y).pathVal == 0 )
            {
                PKGLISTNODE *nextp = p->next;

                free( p );
                if ( prevp )
                {
                    p = prevp->next = nextp;
                }
                else
                {
                    p = gbKnownPkgs = nextp;

                }
            }
            else
            {
                prevp = p;
                p = p->next;
            }
        }

        FindAnyNeighbour( &dir );
    }
    else if ( minPkg )
    {
        /* Otherwise, see where to go next */
        ShortestPath( minPkg->pkg.x, minPkg->pkg.y );

        FindMinNeighbour( &dir );
    }
    else
    {
        FindAnyNeighbour( &dir );
    }

    if ( dir == NODIR )
    {
        IssuePick( 1, NULL );
    }
    else
    {
        IssueMove( bid, dir );
    }
}

void GotoHomeBaseMode( long bid )
{
    long minVal;
    HOMEBASENODE *minHomeBase;
    HOMEBASENODE *p;
    char dir;

#ifdef ICFP_DEBUG
    printf("In GotoHomeBaseMode\n");
#endif

    /* Find nearest home base */
    minVal = LONG_MAX;
    minHomeBase = NULL;

    p = gbHomeBase;

    while ( p != NULL )
    {
        if ( Tile(p->x, p->y).pathVal < minVal )
        {
            minHomeBase = p;
            minVal = Tile(p->x, p->y).pathVal;
        }

        p = p->next;
    }

    /* If its supposed to be here, forget it since I never picked anything */
    if ( minVal == 0 )
    {
        /* we could check whether there are any valid packages, and
        if yes, not delete it (though in rare cases it could lead to
        infinite loop) */
        gbHomeBase = DeleteHomeBaseNode( gbHomeBase, minHomeBase );

        FindAnyNeighbour( &dir );
    }
    else if ( minHomeBase != NULL )
    {
        /* Otherwise, see where to go next */
        ShortestPath( minHomeBase->x, minHomeBase->y );
        FindMinNeighbour( &dir );
    }
    else
    {
        FindAnyNeighbour( &dir );
    }

    if ( dir == NODIR )
    {
        IssuePick( 1, NULL );
    }
    else
    {
        IssueMove( bid, dir );
    }
}


void RandomWalkMode( long bid )
{
    char dir;

#ifdef ICFP_DEBUG
    printf("In RandomWalkMode\n");
#endif

    FindAnyNeighbour( &dir  );

    if ( dir == NODIR )
    {
        IssuePick( 1, NULL );
    }
    else
    {
        IssueMove( bid, dir );
    }
}


void IssueRightCommand( PKGLISTNODE *nowPkgs )
{
    long bid;
    int timeToDrop;
    INTLISTNODE *pkgIDs;

    /* Call ShortestPath to determine path lengths */
    ShortestPath(gbMyX, gbMyY);

    GetBidValue( &bid, &timeToDrop);

    if ( ( bid == 1 ) && ( nowPkgs != NULL ) && !timeToDrop )
    {
        pkgIDs = WhatToPick( nowPkgs ); /* Note: Don't pick unreachable packages */
        if ( pkgIDs != NULL )
        {
            IssuePick( bid, pkgIDs );
            FreeIntList( pkgIDs );
            return;
        }
    }

    if ( gbMyPkgs != NULL )
    {
        PackageDeliveryMode( bid );
        return;
    }
    else if ( gbKnownPkgs != NULL )
    {
        PackagePickMode( bid );
        return;
    }
    else if ( gbHomeBase != NULL )
    {
        GotoHomeBaseMode( bid );
        return;
    }
    else
    {
        RandomWalkMode( bid );
        return;
    }

    IssuePick( 1, NULL );
}


